import unittest

from pathlib import Path
import tempfile
import rdflib
from rdflib.compare import graph_diff

import numpy as np
import quantities as pq
import neo

from alpaca.alpaca_types import (DataObject, File, FunctionInfo,
                                 FunctionExecution,
                                 Container)
from alpaca import AlpacaProvDocument, alpaca_setting
from alpaca.serialization.converters import _ensure_type
from alpaca.serialization.neo import _neo_to_prov

# Define tuples of information as they would be captured by the decorator
# The unit tests will build FunctionExecution tuples using them

# Function
TEST_FUNCTION = FunctionInfo("test_function", "test", "0.0.1")

# Object without metadata
INPUT = DataObject("12345", "joblib_SHA1", "test.InputObject", 12345, {}, None)

# Object with all main types of metadata
INPUT_METADATA = DataObject("12345", "joblib_SHA1", "test.InputObject", 12345,
                            details={'metadata_1': "value1",
                                     'metadata_2': 5,
                                     'metadata_3': 5.0,
                                     'metadata_4': True},
                            value=None)

OUTPUT_METADATA_NEO = DataObject("54321", "joblib_SHA1",
                                 "neo.core.SpikeTrain", 54321,
                                 details={'name': "Spiketrain#1",
                                          'annotations': {'sua': False,
                                                          'channel': 56},
                                          'array_annotations': {
                                              'complexity': np.array(
                                                  [0, 1, 2, 3]),
                                              'event': np.array(
                                                  [True, False, False])}
                                          },
                                 value=None)

# Object with special metadata

# Files
INPUT_FILE = File("56789", "sha256", "/test_file_input")
OUTPUT_FILE = File("98765", "sha256", "/test_file_output")

# Simple objects to test multiple inputs/outputs handling
INPUT_2 = DataObject("212345", "joblib_SHA1", "test.InputObject", 212345, {},
                     None)
OUTPUT = DataObject("54321", "joblib_SHA1", "test.OutputObject", 54321, {},
                    None)
OUTPUT_2 = DataObject("254321", "joblib_SHA1", "test.OutputObject", 254321, {},
                      None)

# None output
NONE_OUTPUT = DataObject("777777", "UUID", "builtins.NoneType", 777777, {},
                         None)

# Object collections
COLLECTION = DataObject("888888", "joblib_SHA1", "builtins.list", 888888, {},
                        None)

# General information. Will be fixed across the tests
TIMESTAMP_START = "2022-05-02T12:34:56.123456"
TIMESTAMP_END = "2022-05-02T12:35:56.123456"
SCRIPT_INFO = File("111111", "sha256", "/script.py")
SCRIPT_SESSION_ID = "999999"


def assert_rdf_graphs_equal(G1, G2):
    result = G1.isomorphic(G2)
    _, in_first, in_second = graph_diff(G1, G2)
    result = result and (len(in_first) == 0)
    result = result and (len(in_second) == 0)
    return result


class AlpacaProvSerializationTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ttl_path = Path(__file__).parent / "res"
        alpaca_setting('authority', "fz-juelich.de")

    def setUp(self):
        alpaca_setting('store_values', [])

    def test_value_serialization(self):
        # DataObject tuples for each type that should be captured
        # They are output of the simulated output
        alpaca_setting('store_values', ['builtins.dict'])

        INT = DataObject("543211", "joblib_SHA1", "builtins.int", 543211,
                         {}, 1)
        FLOAT = DataObject("543212", "joblib_SHA1", "builtins.float", 543212,
                           {}, 1.1)
        STR = DataObject("543213", "joblib_SHA1", "builtins.str", 543213,
                         {}, "test")
        COMPLEX = DataObject("543214", "joblib_SHA1", "builtins.complex",
                             543214, {}, 3+5j)
        BOOL = DataObject("543215", "joblib_SHA1", "builtins.bool", 543215,
                          {}, True)
        NUMPY_FLOAT32 = DataObject("543216", "joblib_SHA1", "numpy.float32",
                                   543216, {}, np.float32(1.2))
        NUMPY_FLOAT64 = DataObject("543217", "joblib_SHA1", "numpy.float64",
                                   543217, {}, np.float64(1.3))
        NUMPY_INT64 = DataObject("543218", "joblib_SHA1", "numpy.int64",
                                 543218, {}, np.int64(2))
        NUMPY_INT32 = DataObject("543219", "joblib_SHA1", "numpy.int32",
                                 543219, {}, np.int32(3))
        NUMPY_INT16 = DataObject("5432110", "joblib_SHA1", "numpy.int16",
                                 5432110, {}, np.int16(-4))

        DICT = DataObject("5432111", "joblib_SHA1", "builtins.dict",
                          5432111, {},
                          str(dict(id=[1, 2, 3], value={4, 5, 6})))

        function_execution = FunctionExecution(
            function=TEST_FUNCTION,
            input={'input_1': INPUT}, params={'param_1': 5},
            output={0: OUTPUT, 1: INT, 2: FLOAT, 3: STR, 4: COMPLEX,
                    5: BOOL, 6: NUMPY_FLOAT32, 7: NUMPY_FLOAT64,
                    8: NUMPY_INT64, 9: NUMPY_INT32, 10: NUMPY_INT16,
                    11: DICT},
            call_ast=None,
            arg_map=['input_1', 'param_1'], kwarg_map=[], return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="test_function(input_1, 5)"
        )

        # Load expected RDF graph
        expected_graph_file = self.ttl_path / "values.ttl"
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected_graph_file, format='turtle')

        # Serialize the history using AlpacaProv document
        alpaca_prov = AlpacaProvDocument()
        alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                history=[function_execution])

        # Check if graphs are equal
        self.assertTrue(assert_rdf_graphs_equal(alpaca_prov.graph,
                                                expected_graph))

    def test_input_output_serialization(self):
        function_execution = FunctionExecution(
            function=TEST_FUNCTION,
            input={'input_1': INPUT}, params={'param_1': 5},
            output={0: OUTPUT}, call_ast=None,
            arg_map=['input_1', 'param_1'], kwarg_map=[], return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="test_function(input_1, 5)"
        )

        # Load expected RDF graph
        expected_graph_file = self.ttl_path / "input_output.ttl"
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected_graph_file, format='turtle')

        # Serialize the history using AlpacaProv document
        alpaca_prov = AlpacaProvDocument()
        alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                history=[function_execution])

        # Check if graphs are equal
        self.assertTrue(assert_rdf_graphs_equal(alpaca_prov.graph,
                                                expected_graph))

    def test_metadata_serialization(self):
        function_execution = FunctionExecution(
            function=TEST_FUNCTION,
            input={'input_1': INPUT_METADATA}, params={'param_1': 5},
            output={0: OUTPUT_METADATA_NEO}, call_ast=None,
            arg_map=['input_1', 'param_1'], kwarg_map=[], return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="test_function(input_1, 5)"
        )

        # Load expected RDF graph
        expected_graph_file = self.ttl_path / "metadata.ttl"
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected_graph_file, format='turtle')

        # Serialize the history using AlpacaProv document
        alpaca_prov = AlpacaProvDocument()
        alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                history=[function_execution])

        # Check if graphs are equal
        self.assertTrue(assert_rdf_graphs_equal(alpaca_prov.graph,
                                                expected_graph))

    def test_input_container_serialization(self):
        function_execution = FunctionExecution(
            function=TEST_FUNCTION,
            input={'input_container': Container((INPUT, INPUT_2))},
            params={'param_1': 5},
            output={0: OUTPUT}, call_ast=None,
            arg_map=['input_container', 'param_1'], kwarg_map=[],
            return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="test_function(input_container, 5)"
        )

        # Load expected RDF graph
        expected_graph_file = self.ttl_path / "input_container.ttl"
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected_graph_file, format='turtle')

        # Serialize the history using AlpacaProv document
        alpaca_prov = AlpacaProvDocument()
        alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                history=[function_execution])

        # Check if graphs are equal
        self.assertTrue(assert_rdf_graphs_equal(alpaca_prov.graph,
                                                expected_graph))

    def test_class_method_serialization(self):
        obj_info = DataObject(
            hash="232323",
            hash_method="joblib_SHA1",
            type="test.ObjectWithMethod",
            id=232323,
            details={}, value=None)

        function_execution = FunctionExecution(
            function=FunctionInfo('ObjectWithMethod.process',
                                  'test', ''),
            input={'self': obj_info, 'array': INPUT},
            params={'param1': 4},
            output={0: OUTPUT}, call_ast=None,
            arg_map=['self', 'array', 'param1'], kwarg_map=[],
            return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="res = obj.process(INPUT, 4)")

        # Load expected RDF graph
        expected_graph_file = self.ttl_path / "class_method.ttl"
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected_graph_file, format='turtle')

        # Serialize the history using AlpacaProv document
        alpaca_prov = AlpacaProvDocument()
        alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                history=[function_execution])

        # Check if graphs are equal
        self.assertTrue(assert_rdf_graphs_equal(alpaca_prov.graph,
                                                expected_graph))

    def test_input_multiple_serialization(self):
        function_execution = FunctionExecution(
            function=TEST_FUNCTION,
            input={'input_1': INPUT, 'input_2': INPUT_2},
            params={'param_1': 5},
            output={0: OUTPUT}, call_ast=None,
            arg_map=['input_1', 'input_2', 'param_1'], kwarg_map=[],
            return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="test_function(input_1, input_2, 5)"
        )

        # Load expected RDF graph
        expected_graph_file = self.ttl_path / "input_multiple.ttl"
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected_graph_file, format='turtle')

        # Serialize the history using AlpacaProv document
        alpaca_prov = AlpacaProvDocument()
        alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                history=[function_execution])

        # Check if graphs are equal
        self.assertTrue(assert_rdf_graphs_equal(alpaca_prov.graph,
                                                expected_graph))

    def test_collection_serialization(self):
        indexing_access = FunctionExecution(
            function=FunctionInfo(name='subscript', module="", version=""),
            input={0: COLLECTION}, params={'index': 0},
            output={0: INPUT}, call_ast=None, arg_map=None, kwarg_map=None,
            return_targets=[], time_stamp_start=TIMESTAMP_START,
            time_stamp_end=TIMESTAMP_END, execution_id="888888", order=None,
            code_statement=None)

        function_execution = FunctionExecution(
            function=TEST_FUNCTION,
            input={'input_1': INPUT}, params={'param_1': 5},
            output={0: OUTPUT}, call_ast=None,
            arg_map=['input_1', 'param_1'], kwarg_map=[], return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="test_function(source_list[0], 5)"
        )

        # Load expected RDF graph
        expected_graph_file = self.ttl_path / "collection.ttl"
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected_graph_file, format='turtle')

        # Serialize the history using AlpacaProv document
        alpaca_prov = AlpacaProvDocument()
        alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                history=[indexing_access, function_execution])

        # Check if graphs are equal
        self.assertTrue(assert_rdf_graphs_equal(alpaca_prov.graph,
                                                expected_graph))

    def test_file_output_serialization(self):
        function_execution = FunctionExecution(
            function=TEST_FUNCTION,
            input={'input_1': INPUT}, params={'param_1': 5},
            output={0: NONE_OUTPUT, 'file.0': OUTPUT_FILE}, call_ast=None,
            arg_map=['input_1', 'param_1'], kwarg_map=[], return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="test_function(input_1, 5)"
        )

        # Load expected RDF graph
        expected_graph_file = self.ttl_path / "file_output.ttl"
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected_graph_file, format='turtle')

        # Serialize the history using AlpacaProv document
        alpaca_prov = AlpacaProvDocument()
        alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                history=[function_execution])

        # Check if graphs are equal
        self.assertTrue(assert_rdf_graphs_equal(alpaca_prov.graph,
                                                expected_graph))

    def test_file_input_serialization(self):
        function_execution = FunctionExecution(
            function=TEST_FUNCTION,
            input={'input_1': INPUT_FILE}, params={'param_1': 5},
            output={0: OUTPUT}, call_ast=None,
            arg_map=['input_1', 'param_1'], kwarg_map=[], return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="test_function(input_1, 5)"
        )

        # Load expected RDF graph
        expected_graph_file = self.ttl_path / "file_input.ttl"
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected_graph_file, format='turtle')

        # Serialize the history using AlpacaProv document
        alpaca_prov = AlpacaProvDocument()
        alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                history=[function_execution])

        # Check if graphs are equal
        self.assertTrue(assert_rdf_graphs_equal(alpaca_prov.graph,
                                                expected_graph))


class SerializationIOTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.ttl_path = Path(__file__).parent / "res"
        cls.temp_dir = tempfile.TemporaryDirectory(dir=cls.ttl_path,
                                                   suffix="tmp")
        function_execution = FunctionExecution(
            function=TEST_FUNCTION,
            input={'input_1': INPUT}, params={'param_1': 5},
            output={0: OUTPUT}, call_ast=None,
            arg_map=['input_1'], kwarg_map=[], return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="test_function(input_1, 5)"
        )

        # Serialize the history using AlpacaProv document
        cls.alpaca_prov = AlpacaProvDocument()
        cls.alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                    history=[function_execution])
        alpaca_setting('authority', "fz-juelich.de")

    def test_serialization_deserialization(self):

        # For every supported format, serialize to a temp file
        for output_format in ('json-ld', 'n3', 'nt', 'hext', 'pretty-xml',
                              'trig', 'turtle', 'longturtle', 'xml'):
            with self.subTest(f"Serialization format",
                              output_format=output_format):
                output_file = Path(
                    self.temp_dir.name) / f"test.{output_format}"
                self.alpaca_prov.serialize(output_file,
                                           file_format=output_format)
                self.assertTrue(output_file.exists())

        # For every supported format with parsers, read the temp saved files
        # and check against the original graph.
        # No parsers for: 'pretty-xml', 'longturtle', 'hext' and 'trig'
        for read_format in ('json-ld', 'n3', 'nt', 'turtle', 'xml'):
            with self.subTest(f"Deserialization format",
                              read_format=read_format):
                input_file = Path(self.temp_dir.name) / f"test.{read_format}"
                read_alpaca_prov = AlpacaProvDocument()
                read_alpaca_prov.read_records(input_file, file_format=None)
                self.assertTrue(assert_rdf_graphs_equal(self.alpaca_prov.graph,
                                                        read_alpaca_prov.graph))

        # Test unsupported formats
        for wrong_format in ('pretty-xml', 'longturtle', 'hext', 'trig'):
            with self.subTest(f"Unsupported format",
                              wrong_format=wrong_format):
                with self.assertRaises(ValueError):
                    input_file = Path(
                        self.temp_dir.name) / f"test.{wrong_format}"
                    read_alpaca_prov = AlpacaProvDocument()
                    read_alpaca_prov.read_records(input_file, file_format=None)

    def test_shortcut_format(self):
        input_ttl = self.ttl_path / "input_output.ttl"
        read_ttl = AlpacaProvDocument()
        read_ttl.read_records(input_ttl, file_format=None)
        self.assertTrue(assert_rdf_graphs_equal(read_ttl.graph,
                                                self.alpaca_prov.graph))

    def test_no_format(self):
        no_ext = Path(self.temp_dir.name) / "no_ext"
        self.alpaca_prov.serialize(no_ext, file_format='turtle')
        read_no_ext = AlpacaProvDocument()
        with self.assertRaises(ValueError):
            read_no_ext.read_records(no_ext, file_format=None)


class ConvertersTestCase(unittest.TestCase):

    def test_list(self):
        obj = [1, 2, 3]
        expected = "[1, 2, 3]"
        self.assertEqual(expected, _ensure_type(obj))

    def test_quantity(self):
        obj = 1 * pq.ms
        expected = "1.0 ms"
        self.assertEqual(expected, _ensure_type(obj))

    def test_neo(self):
        segment = neo.Segment(name="test")
        spiketrain = neo.SpikeTrain([1, 2, 3] * pq.ms, t_stop=10 * pq.ms)
        segment.spiketrains = [spiketrain]

        expected = "neo.core.segment.Segment(t_start=0.0 ms, " \
                   "t_stop=10.0 ms, name=test, description=None)"
        segment_str = _ensure_type(segment)
        self.assertEqual(expected, segment_str)

    def test_neo_selected_attributes(self):
        segment = neo.Segment(name="test")
        spiketrain = neo.SpikeTrain([1, 2, 3] * pq.ms, t_stop=10 * pq.ms)
        segment.spiketrains = [spiketrain]

        expected = "neo.core.segment.Segment(name=test, t_stop=10.0 ms)"
        segment_str = _neo_to_prov(segment, ['name', 't_stop'])
        self.assertEqual(expected, segment_str)

    def test_neo_single_attributes(self):
        segment = neo.Segment(name="test")

        expected = "neo.core.segment.Segment(name=test)"
        segment_str = _neo_to_prov(segment, ['name'])
        self.assertEqual(expected, segment_str)

    def test_base_types(self):
        self.assertEqual(True, _ensure_type(True))
        self.assertEqual(1, _ensure_type(1))
        self.assertEqual(1.56, _ensure_type(1.56))
        self.assertEqual("test", _ensure_type("test"))

    def test_others(self):
        value = 1.0 + 5.0j
        self.assertEqual("(1+5j)", _ensure_type(value))


class MultipleMembershipSerializationTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        alpaca_setting('authority', "fz-juelich.de")

    def test_multiple_memberships(self):
        # test relationship `super_container.containers[0].inputs[1]`
        self.ttl_path = Path(__file__).parent / "res"

        super_container = DataObject("2333333", "joblib_SHA1",
                                     "test.SuperContainer", 2333333, {}, None)

        super_container_list = DataObject("23333332", "joblib_SHA1",
                                          "builtins.list", 23333332, {}, None)

        container = DataObject("333333", "joblib_SHA1", "test.Container", 333333,
                               {}, None)

        container_list = DataObject("3333332", "joblib_SHA1", "builtins.list",
                                    3333332, {}, None)

        attribute_access_container = FunctionExecution(
            function=FunctionInfo(name='attribute', module="", version=""),
            input={0: super_container}, params={'name': 'containers'},
            output={0: super_container_list}, call_ast=None, arg_map=None,
            kwarg_map=None,
            return_targets=[], time_stamp_start=TIMESTAMP_START,
            time_stamp_end=TIMESTAMP_END, execution_id="888888",
            order=None, code_statement=None)

        indexing_access_container = FunctionExecution(
            function=FunctionInfo(name='subscript', module="", version=""),
            input={0: super_container_list}, params={'index': 0},
            output={0: container}, call_ast=None, arg_map=None, kwarg_map=None,
            return_targets=[], time_stamp_start=TIMESTAMP_START,
            time_stamp_end=TIMESTAMP_END, execution_id="2888888",
            order=None,
            code_statement=None)

        attribute_access_inputs = FunctionExecution(
            function=FunctionInfo(name='attribute', module="", version=""),
            input={0: container}, params={'name': 'inputs'},
            output={0: container_list}, call_ast=None, arg_map=None,
            kwarg_map=None,
            return_targets=[], time_stamp_start=TIMESTAMP_START,
            time_stamp_end=TIMESTAMP_END, execution_id="3888888",
            order=None, code_statement=None)

        indexing_access_inputs = FunctionExecution(
            function=FunctionInfo(name='subscript', module="", version=""),
            input={0: container_list}, params={'index': 1},
            output={0: INPUT}, call_ast=None, arg_map=None, kwarg_map=None,
            return_targets=[], time_stamp_start=TIMESTAMP_START,
            time_stamp_end=TIMESTAMP_END, execution_id="4888888",
            order=None,
            code_statement=None)

        function_execution = FunctionExecution(
            function=TEST_FUNCTION,
            input={'input_1': INPUT}, params={'param_1': 5},
            output={0: OUTPUT}, call_ast=None,
            arg_map=['input_1', 'param_1'], kwarg_map=[],
            return_targets=[],
            time_stamp_start=TIMESTAMP_START, time_stamp_end=TIMESTAMP_END,
            execution_id="12345", order=1,
            code_statement="test_function(super_container.containers[0].inputs[1], 5)"
        )

        # Load expected RDF graph
        expected_graph_file = self.ttl_path / "multiple_memberships.ttl"
        expected_graph = rdflib.Graph()
        expected_graph.parse(expected_graph_file, format='turtle')

        # Serialize the history using AlpacaProv document
        alpaca_prov = AlpacaProvDocument()
        alpaca_prov.add_history(SCRIPT_INFO, SCRIPT_SESSION_ID,
                                history=[attribute_access_container,
                                         indexing_access_container,
                                         attribute_access_inputs,
                                         indexing_access_inputs,
                                         function_execution])

        # Check if graphs are equal
        self.assertTrue(assert_rdf_graphs_equal(alpaca_prov.graph,
                                                expected_graph))


class NeoMetadataPluginTestCase(unittest.TestCase):
    pass


if __name__ == "__main__":
    unittest.main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
from datetime import date


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'Alpaca'
authors = 'Alpaca authors and contributors'
copyright = u"2022-{this_year}, Forschungszentrum Jülich GmbH, INM-6, IAS-6."\
    "All rights reserved.".format(this_year=date.today().year,)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
root_dir = os.path.dirname(os.path.dirname(__file__))
with open(os.path.join(root_dir, 'alpaca', 'VERSION')) as version_file:
    # The full version, including alpha/beta/rc tags.
    release = version_file.read().strip()

# The short X.Y version.
version = '.'.join(release.split('.')[:-1])


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.viewcode',
    'sphinx_tabs.tabs',
    'numpydoc',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build',
                    '**.ipynb_checkpoints',
                    'Thumbs.db',
                    '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# Required to automatically create a summary page for each function listed in
# the autosummary fields of each module.
autosummary_generate = True

# Set to False to not overwrite our custom toctree/*.rst
autosummary_generate_overwrite = True


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

html_theme_options = {
    'font_family': 'Arial',
    'page_width': '1200px',  # default is 940
    'sidebar_width': '280px',  # default is 220
    'logo': 'images/alpaca_logo.png',
}

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = 'images/*.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# CSS file to customize and tweak display
html_css_files = [
    'css/style.css',
]


# Suppresses wrong numpy doc warnings
# see here https://github.com/phn/pytpm/issues/3#issuecomment-12133978
numpydoc_show_class_members = False

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = False

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# Output file base name for HTML help builder.
htmlhelp_basename = 'alpacadoc'



import os

from setuptools import setup

with open(os.path.join(os.path.dirname(__file__),
                       "alpaca", "VERSION")) as version_file:
    version = version_file.read().strip()

with open("README.md") as f:
    long_description = f.read()

with open('requirements/requirements.txt') as fp:
    install_requires = fp.read()


setup(
    name="alpaca-prov",
    version=version,
    packages=['alpaca', 'alpaca.utils', 'alpaca.serialization',
              'alpaca.ontology', 'alpaca.code_analysis'],
    include_package_data=True,
    install_requires=install_requires,
    author="Alpaca authors and contributors",
    author_email="",
    description="Alpaca is a Python package for the capture of provenance "
                "information during the execution of Python scripts that "
                "process data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="BSD",
    url='https://github.com/INM-6/alpaca',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering']
)



import unittest
import io
from rdflib import Literal, URIRef, Namespace, Graph, RDF, PROV

from alpaca import activate, deactivate, Provenance, save_provenance
from alpaca.ontology.annotation import (_OntologyInformation,
                                        ONTOLOGY_INFORMATION)
from alpaca.ontology import ALPACA

from collections import Counter

# Ontology namespace definition used for the tests
EXAMPLE_NS = {'ontology': "http://example.org/ontology#"}


##############################
# Test objects to be annotated
##############################

class InputObject:
    __ontology__ = {
        "data_object": "ontology:InputObject",
        "namespaces": EXAMPLE_NS}


class OutputObject:
    __ontology__ = {
        "data_object": "ontology:OutputObject",
        "attributes": {'name': "ontology:Attribute"},
        "namespaces": EXAMPLE_NS}

    def __init__(self, name, channel):
        self.name = name
        self.channel = channel


class InputObjectURI:
    __ontology__ = {"data_object": "<http://purl.org/ontology#InputObject>"}


#######################################################
# Test functions to be annotated and provenance tracked
#######################################################

@Provenance(inputs=['input'])
def process(input, param_1):
    return OutputObject("SpikeTrain#1", 45)

process.__wrapped__.__ontology__ = {
    "function": "ontology:ProcessFunction",
    "namespaces": EXAMPLE_NS,
    "arguments": {'param_1': "ontology:Parameter"},
    "returns": {0: "ontology:ProcessedData"}
}


@Provenance(inputs=['input'])
def process_one_and_process_two(input, param_1):
    return OutputObject("SpikeTrain#1", 45)

process_one_and_process_two.__wrapped__.__ontology__ = {
    "function": ["ontology:Process1Function", "ontology:Process2Function"],
    "namespaces": EXAMPLE_NS,
    "arguments": {'param_1': "ontology:Parameter"},
    "returns": {0: "ontology:ProcessedData"}
}


@Provenance(inputs=['input'])
def process_multiple(input, param_1):
    return "not_annotated", OutputObject("SpikeTrain#2", 34)

process_multiple.__wrapped__.__ontology__ = {
    "function": "ontology:ProcessFunctionMultiple",
    "namespaces": EXAMPLE_NS,
    "arguments": {'param_1': "ontology:Parameter"},
    "returns": {1: "ontology:ProcessedDataMultiple"}
}


@Provenance(inputs=[], container_output=True)
def process_container_output():
    return list(range(3))

process_container_output.__wrapped__.__ontology__ = {
    "function": "ontology:ProcessContainerOutput",
    "namespaces": EXAMPLE_NS,
    "returns": {'*': "ontology:ProcessedContainerOutput"}
}


@Provenance(inputs=[], container_output=1)
def process_multiple_container_output():
    return [list(range(i, i + 3)) for i in range(0, 7, 3)]

process_multiple_container_output.__wrapped__.__ontology__ = {
    "function": "ontology:ProcessMultipleContainerOutput",
    "namespaces": EXAMPLE_NS,
    "returns": {'***': "ontology:ProcessedMultipleContainerOutput"}
}


@Provenance(inputs=[], container_output=1)
def process_multiple_container_output_multiple_annotations():
    return [list(range(i, i + 3)) for i in range(0, 4, 3)]

process_multiple_container_output_multiple_annotations.__wrapped__.__ontology__ = {
    "function": "ontology:ProcessMultipleContainerOutputMultipleAnnotations",
    "namespaces": EXAMPLE_NS,
    "returns": {
        '**': "ontology:ProcessedMultipleContainerOutputLevel1",
        '***': "ontology:ProcessedMultipleContainerOutputLevel2"}
}


@Provenance(inputs=[], container_output=1)
def process_multiple_container_output_multiple_annotations_root():
    return [list(range(i, i + 3)) for i in range(0, 4, 3)]

process_multiple_container_output_multiple_annotations_root.__wrapped__.__ontology__ = {
    "function": "ontology:ProcessMultipleContainerOutputMultipleAnnotationsRoot",
    "namespaces": EXAMPLE_NS,
    "returns": {
        '*': "ontology:ProcessedMultipleContainerOutputLevel0",
        '**': "ontology:ProcessedMultipleContainerOutputLevel1",
        '***': "ontology:ProcessedMultipleContainerOutputLevel2"}
}


@Provenance(inputs=[], container_output=(1, 1))
def process_multiple_container_output_multiple_annotations_range():
    return [list(range(i, i + 3)) for i in range(0, 4, 3)]

process_multiple_container_output_multiple_annotations_range.__wrapped__.__ontology__ = {
    "function": "ontology:ProcessMultipleContainerOutputMultipleAnnotationsRange",
    "namespaces": EXAMPLE_NS,
    "returns": {
        '*': "ontology:ProcessedMultipleContainerOutputLevel0",
        '**': "ontology:ProcessedMultipleContainerOutputLevel1",
        '***': "ontology:ProcessedMultipleContainerOutputLevel2"}
}


@Provenance(inputs=['input'])
def process_input_annotation(input, param):
    return input + 2

process_input_annotation.__wrapped__.__ontology__ = {
    "function": "ontology:ProcessInputAnnotation",
    "namespaces": EXAMPLE_NS,
    "arguments": {'input': "ontology:Input", 'param': "ontology:Param"}
}


@Provenance(inputs=['input'], container_input=['input_list'])
def process_container_input_annotation(input, input_list, param):
    return [i + input for i in input_list]

process_container_input_annotation.__wrapped__.__ontology__ = {
    "function": "ontology:ProcessContainerInputAnnotation",
    "namespaces": EXAMPLE_NS,
    "arguments": {'input': "ontology:Input",
                  'input_list': "ontology:ContainerElementInput",
                  'param': "ontology:Param"}
}


@Provenance(inputs=['input'])
def process_no_annotations(input):
    return input + 2


############
# Unit tests
############

class OntologyAnnotationTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Create rdflib Namespace for tests
        cls.ONTOLOGY = Namespace(EXAMPLE_NS['ontology'])

    def setUp(self):
        _OntologyInformation.namespaces.clear()
        ONTOLOGY_INFORMATION.clear()

    def test_redefine_namespaces(self):
        obj = InputObject()
        self.assertDictEqual(_OntologyInformation.namespaces, {})
        info = _OntologyInformation(obj)

        # At this point, the class should be updated with the 'ontology'
        # namespace
        with self.assertRaises(ValueError):
            info.add_namespace('ontology', "http://purl.org/ontology")

        info.add_namespace('purl_ontology', "http://purl.org/ontology")
        self.assertEqual(len(info.namespaces), 2)

        self.assertEqual(info.namespaces['ontology'], self.ONTOLOGY)
        self.assertEqual(info.namespaces['purl_ontology'],
                         Namespace("http://purl.org/ontology"))

    def test_annotation_object_input_uri(self):
        obj = InputObjectURI()
        self.assertIsNotNone(
            _OntologyInformation.get_ontology_information(obj))
        info = _OntologyInformation(obj)
        self.assertEqual(
            info.get_uri("data_object"),
            URIRef("http://purl.org/ontology#InputObject"))

        # Namespaces included in representation as this is a class attribute
        self.assertEqual(
            str(info),
            "OntologyInformation(data_object='"
            "<http://purl.org/ontology#InputObject>', namespaces={})"
        )

    def test_annotation_object_input(self):
        obj = InputObject()
        self.assertIsNotNone(
            _OntologyInformation.get_ontology_information(obj))
        info = _OntologyInformation(obj)
        self.assertEqual(
            info.get_uri("data_object"),
            URIRef("http://example.org/ontology#InputObject"))
        self.assertEqual(
            str(info),
            "OntologyInformation(data_object='ontology:InputObject', "
            f"namespaces={{'ontology': {repr(self.ONTOLOGY)}}})"
        )

    def test_annotation_object_output(self):
        obj = OutputObject("test", 45)
        self.assertIsNotNone(
            _OntologyInformation.get_ontology_information(obj))
        info = _OntologyInformation(obj)
        self.assertEqual(
            info.get_uri("data_object"),
            URIRef("http://example.org/ontology#OutputObject"))
        self.assertEqual(
            info.get_uri("attributes", "name"),
            URIRef("http://example.org/ontology#Attribute"))
        self.assertEqual(
            str(info),
            "OntologyInformation(data_object='ontology:OutputObject', "
            "attributes={'name': 'ontology:Attribute'}, "
            f"namespaces={{'ontology': {repr(self.ONTOLOGY)}}})"
        )

    def test_annotation_function(self):
        self.assertIsNotNone(
            _OntologyInformation.get_ontology_information(process))
        info = _OntologyInformation(process)
        self.assertEqual(
            info.get_uri("function"),
            URIRef("http://example.org/ontology#ProcessFunction"))
        self.assertEqual(
            info.get_uri("arguments", "param_1"),
            URIRef("http://example.org/ontology#Parameter"))
        self.assertEqual(
            info.get_uri("returns", 0),
            URIRef("http://example.org/ontology#ProcessedData"))
        self.assertEqual(
            str(info),
            "OntologyInformation(function='ontology:ProcessFunction', "
            "arguments={'param_1': 'ontology:Parameter'}, "
            f"namespaces={{'ontology': {repr(self.ONTOLOGY)}}}, "
            "returns={0: 'ontology:ProcessedData'})"
        )

    def test_annotation_function_multiple_annotations(self):
        self.assertIsNotNone(
            _OntologyInformation.get_ontology_information(
                process_one_and_process_two))
        info = _OntologyInformation(process_one_and_process_two)
        self.assertListEqual(
            info.get_uri("function"),
            [URIRef("http://example.org/ontology#Process1Function"),
             URIRef("http://example.org/ontology#Process2Function")])
        self.assertEqual(
            info.get_uri("arguments", "param_1"),
            URIRef("http://example.org/ontology#Parameter"))
        self.assertEqual(
            info.get_uri("returns", 0),
            URIRef("http://example.org/ontology#ProcessedData"))
        self.assertEqual(
            str(info),
            "OntologyInformation(function='['ontology:Process1Function', "
            "'ontology:Process2Function']', "
            "arguments={'param_1': 'ontology:Parameter'}, "
            f"namespaces={{'ontology': {repr(self.ONTOLOGY)}}}, "
            "returns={0: 'ontology:ProcessedData'})")

    def test_annotation_function_multiple(self):
        self.assertIsNotNone(
            _OntologyInformation.get_ontology_information(process_multiple))
        info = _OntologyInformation(process_multiple)
        self.assertEqual(
            info.get_uri("function"),
            URIRef("http://example.org/ontology#ProcessFunctionMultiple"))
        self.assertEqual(
            info.get_uri("arguments", "param_1"),
            URIRef("http://example.org/ontology#Parameter"))
        self.assertEqual(
            info.get_uri("returns", 1),
            URIRef("http://example.org/ontology#ProcessedDataMultiple"))
        self.assertEqual(
            str(info),
            "OntologyInformation(function='ontology:ProcessFunctionMultiple', "
            "arguments={'param_1': 'ontology:Parameter'}, "
            f"namespaces={{'ontology': {repr(self.ONTOLOGY)}}}, "
            "returns={1: 'ontology:ProcessedDataMultiple'})"
        )

    def test_invalid_object_annotations(self):
        obj = InputObject()
        info = _OntologyInformation(obj)
        self.assertIsNone(info.get_uri("attributes", "name"))
        self.assertIsNone(info.get_uri("attributes", "channel"))
        self.assertIsNone(info.get_uri("non_existent"))
        self.assertIsNone(info.get_uri("non_existent", "test"))

        output_obj = OutputObject("test", 45)
        output_info = _OntologyInformation(output_obj)
        self.assertIsNotNone(output_info.get_uri("attributes", "name"))
        self.assertIsNone(output_info.get_uri("attributes", "channel"))
        self.assertIsNone(output_info.get_uri("non_existent"))
        self.assertIsNone(output_info.get_uri("non_existent", "test"))

    def test_namespaces(self):
        input_obj = InputObject()
        output_obj = OutputObject("test", 45)

        input_info = _OntologyInformation(input_obj)
        output_info = _OntologyInformation(output_obj)
        function_info = _OntologyInformation(process)

        for info in (input_info, output_info, function_info):
            self.assertEqual(info.namespaces['ontology'], self.ONTOLOGY)
            self.assertTupleEqual(tuple(info.namespaces.keys()), ('ontology',))

    def test_provenance_no_annotation(self):
        activate(clear=True)
        result = process_no_annotations(5)
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that no other annotations are present
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        types = list(prov_graph.objects(execution_uri, RDF.type))
        self.assertListEqual(types, [ALPACA.FunctionExecution])

    def test_provenance_annotation(self):
        activate(clear=True)
        input_object = InputObject()
        output_object = process(input_object, 34)
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that the annotations exist (1 per class is expected)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.Parameter)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessFunction)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedData)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.InputObject)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.OutputObject)))
            ), 1)

        # FunctionExecution is ProcessFunction
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        self.assertTrue((execution_uri,
                         RDF.type,
                         self.ONTOLOGY.ProcessFunction) in prov_graph)

        # Check parameter name
        parameter_node = list(
            prov_graph.subjects(RDF.type, self.ONTOLOGY.Parameter))[0]
        self.assertTrue((parameter_node,
                         ALPACA.pairName, Literal("param_1")) in prov_graph)
        self.assertTrue((parameter_node,
                         ALPACA.pairValue, Literal(34)) in prov_graph)

        # Check returned value
        output_node = list(
            prov_graph.subjects(RDF.type, self.ONTOLOGY.ProcessedData))[0]
        self.assertTrue((output_node,
                         PROV.wasGeneratedBy, execution_uri) in prov_graph)
        self.assertTrue((output_node,
                         RDF.type, ALPACA.DataObjectEntity) in prov_graph)
        self.assertTrue((output_node,
                         RDF.type, self.ONTOLOGY.OutputObject) in prov_graph)

        # Check attributes of returned value
        expected_attributes = {
            'name': "SpikeTrain#1",
            'channel': 45,
        }
        for attribute in prov_graph.objects(output_node, ALPACA.hasAttribute):
            name = prov_graph.value(attribute, ALPACA.pairName).toPython()
            value = prov_graph.value(attribute, ALPACA.pairValue).toPython()
            self.assertEqual(value, expected_attributes[name])

            # Check if attribute annotation is present for `name`
            if name == 'name':
                self.assertTrue((attribute, RDF.type, self.ONTOLOGY.Attribute)
                                in prov_graph)

        # Check input value
        input_node = list(
            prov_graph.subjects(RDF.type, self.ONTOLOGY.InputObject))[0]
        self.assertTrue((execution_uri, PROV.used, input_node) in prov_graph)
        self.assertTrue((input_node,
                         RDF.type, ALPACA.DataObjectEntity) in prov_graph)
        self.assertTrue((output_node,
                         PROV.wasDerivedFrom, input_node) in prov_graph)

    def test_provenance_multiple_annotations(self):
        activate(clear=True)
        input_object = InputObject()
        output_object = process_one_and_process_two(input_object, 34)
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that the annotations exist (1 per class is expected)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.Parameter)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.Process1Function)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.Process2Function)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedData)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.InputObject)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.OutputObject)))
            ), 1)

        # FunctionExecution is ProcessFunction
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        self.assertTrue((execution_uri,
                         RDF.type,
                         self.ONTOLOGY.Process1Function) in prov_graph)

        self.assertTrue((execution_uri,
                         RDF.type,
                         self.ONTOLOGY.Process2Function) in prov_graph)

        # Check parameter name
        parameter_node = list(
            prov_graph.subjects(RDF.type, self.ONTOLOGY.Parameter))[0]
        self.assertTrue((parameter_node,
                         ALPACA.pairName, Literal("param_1")) in prov_graph)
        self.assertTrue((parameter_node,
                         ALPACA.pairValue, Literal(34)) in prov_graph)

        # Check returned value
        output_node = list(
            prov_graph.subjects(RDF.type, self.ONTOLOGY.ProcessedData))[0]
        self.assertTrue((output_node,
                         PROV.wasGeneratedBy, execution_uri) in prov_graph)
        self.assertTrue((output_node,
                         RDF.type, ALPACA.DataObjectEntity) in prov_graph)
        self.assertTrue((output_node,
                         RDF.type, self.ONTOLOGY.OutputObject) in prov_graph)

        # Check attributes of returned value
        expected_attributes = {
            'name': "SpikeTrain#1",
            'channel': 45,
        }
        for attribute in prov_graph.objects(output_node, ALPACA.hasAttribute):
            name = prov_graph.value(attribute, ALPACA.pairName).toPython()
            value = prov_graph.value(attribute, ALPACA.pairValue).toPython()
            self.assertEqual(value, expected_attributes[name])

            # Check if attribute annotation is present for `name`
            if name == 'name':
                self.assertTrue((attribute, RDF.type, self.ONTOLOGY.Attribute)
                                in prov_graph)

        # Check input value
        input_node = list(
            prov_graph.subjects(RDF.type, self.ONTOLOGY.InputObject))[0]
        self.assertTrue((execution_uri, PROV.used, input_node) in prov_graph)
        self.assertTrue((input_node,
                         RDF.type, ALPACA.DataObjectEntity) in prov_graph)
        self.assertTrue((output_node,
                         PROV.wasDerivedFrom, input_node) in prov_graph)

    def test_provenance_annotation_multiple_returns(self):
        activate(clear=True)
        input_object = InputObject()
        name, output_object = process_multiple(input_object, 45)
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that the annotations exist (1 per class is expected. None
        # are expected for the classes of `process`)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.Parameter)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessFunctionMultiple)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedDataMultiple)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.InputObject)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.OutputObject)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessFunction)))
            ), 0)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedData)))
            ), 0)

        # FunctionExecution is ProcessFunctionMultiple
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        self.assertTrue((execution_uri, RDF.type,
                         self.ONTOLOGY.ProcessFunctionMultiple) in prov_graph)

        # Check parameter name
        parameter_node = list(
            prov_graph.subjects(RDF.type, self.ONTOLOGY.Parameter))[0]
        self.assertTrue((parameter_node,
                         ALPACA.pairName, Literal("param_1")) in prov_graph)
        self.assertTrue((parameter_node,
                         ALPACA.pairValue, Literal(45)) in prov_graph)

        # Check returned value
        output_node = list(
            prov_graph.subjects(RDF.type,
                                self.ONTOLOGY.ProcessedDataMultiple))[0]
        self.assertTrue((output_node,
                         PROV.wasGeneratedBy, execution_uri) in prov_graph)
        self.assertTrue((output_node,
                         RDF.type, ALPACA.DataObjectEntity) in prov_graph)
        self.assertTrue((output_node,
                         RDF.type, self.ONTOLOGY.OutputObject) in prov_graph)

        # Check attributes of returned value
        expected_attributes = {
            'name': "SpikeTrain#2",
            'channel': 34,
        }
        for attribute in prov_graph.objects(output_node, ALPACA.hasAttribute):
            name = prov_graph.value(attribute, ALPACA.pairName).toPython()
            value = prov_graph.value(attribute, ALPACA.pairValue).toPython()
            self.assertEqual(value, expected_attributes[name])

            # Check if attribute annotation is present for `name`
            if name == 'name':
                self.assertTrue((attribute, RDF.type, self.ONTOLOGY.Attribute)
                                in prov_graph)

        # Check input value
        input_node = list(
            prov_graph.subjects(RDF.type, self.ONTOLOGY.InputObject))[0]
        self.assertTrue((execution_uri,
                         PROV.used, input_node) in prov_graph)
        self.assertTrue((input_node,
                         RDF.type, ALPACA.DataObjectEntity) in prov_graph)
        self.assertTrue((output_node,
                         PROV.wasDerivedFrom, input_node) in prov_graph)

    def test_provenance_annotation_container_output(self):
        activate(clear=True)
        container = process_container_output()
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that the annotations exist (1 per class is expected. None
        # are expected for the classes of `process`)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessContainerOutput)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedContainerOutput)))
            ), 3)

        # FunctionExecution is ProcessContainerOutput
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        self.assertTrue((execution_uri, RDF.type,
                         self.ONTOLOGY.ProcessContainerOutput) in prov_graph)

        # Check returned values
        output_nodes = prov_graph.subjects(RDF.type,
                                self.ONTOLOGY.ProcessedContainerOutput)
        for output_node in output_nodes:
            self.assertTrue((output_node,
                             PROV.wasGeneratedBy, execution_uri) in prov_graph)
            self.assertTrue((output_node,
                             RDF.type, ALPACA.DataObjectEntity) in prov_graph)
            self.assertTrue((output_node,
                             RDF.type, self.ONTOLOGY.ProcessedContainerOutput)
                            in prov_graph)

    def test_provenance_annotation_container_multiple_output(self):
        activate(clear=True)
        container = process_multiple_container_output()
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that the annotations exist
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessMultipleContainerOutput)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutput)))
            ), 9)

        # FunctionExecution is ProcessMultipleContainerOutput
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        self.assertTrue((execution_uri, RDF.type,
                         self.ONTOLOGY.ProcessMultipleContainerOutput) in prov_graph)

        # Check returned values
        output_nodes = prov_graph.subjects(RDF.type,
                                self.ONTOLOGY.ProcessedMultipleContainerOutput)
        for output_node in output_nodes:
            self.assertTrue((None, PROV.hadMember, output_node) in prov_graph)
            self.assertTrue((output_node,
                             RDF.type, ALPACA.DataObjectEntity) in prov_graph)
            self.assertTrue((output_node,
                             RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutput)
                            in prov_graph)
            members = list(prov_graph.objects(output_node, PROV.hadMember))
            self.assertEqual(len(members), 0)

    def test_provenance_annotation_container_multiple_output_multiple_annotations(self):
        activate(clear=True)
        container = process_multiple_container_output_multiple_annotations()
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that the annotations exist
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessMultipleContainerOutputMultipleAnnotations)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel2)))
            ), 6)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel1)))
            ), 2)

        # FunctionExecution is ProcessMultipleContainerOutputMultipleAnnotations
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        self.assertTrue((execution_uri, RDF.type,
                         self.ONTOLOGY.ProcessMultipleContainerOutputMultipleAnnotations) in prov_graph)

        # Check returned values
        output_nodes = prov_graph.subjects(RDF.type,
                                self.ONTOLOGY.ProcessedMultipleContainerOutputLevel1)
        for output_node in output_nodes:
            self.assertTrue((None, PROV.hadMember, output_node) in prov_graph)
            self.assertTrue((output_node,
                             RDF.type, ALPACA.DataObjectEntity) in prov_graph)
            self.assertTrue((output_node,
                             RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel1)
                            in prov_graph)
            members = list(prov_graph.objects(output_node, PROV.hadMember))
            self.assertEqual(len(members), 3)
            for element in prov_graph.objects(output_node, PROV.hadMember):
                self.assertTrue(
                    (element, RDF.type, ALPACA.DataObjectEntity) in prov_graph)
                self.assertTrue(
                    (element, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel2) in prov_graph)
                members = list(prov_graph.objects(element, PROV.hadMember))
                self.assertEqual(len(members), 0)

    def test_provenance_annotation_container_multiple_output_multiple_annotations_root(self):
        activate(clear=True)
        container = process_multiple_container_output_multiple_annotations_root()
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that the annotations exist
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessMultipleContainerOutputMultipleAnnotationsRoot)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel2)))
            ), 6)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel1)))
            ), 2)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel0)))
            ), 1)

        # FunctionExecution is ProcessMultipleContainerOutputMultipleAnnotationsRoot
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        self.assertTrue((execution_uri, RDF.type,
                         self.ONTOLOGY.ProcessMultipleContainerOutputMultipleAnnotationsRoot) in prov_graph)

        # Check returned values
        output_nodes = prov_graph.subjects(RDF.type,
                                self.ONTOLOGY.ProcessedMultipleContainerOutputLevel0)
        for output_level0 in output_nodes:
            self.assertTrue((output_level0, PROV.wasGeneratedBy, execution_uri) in prov_graph)
            self.assertTrue((output_level0,
                             RDF.type, ALPACA.DataObjectEntity) in prov_graph)
            self.assertTrue((output_level0,
                             RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel0)
                            in prov_graph)
            members = list(prov_graph.objects(output_level0, PROV.hadMember))
            self.assertEqual(len(members), 2)
            for output_level1 in prov_graph.objects(output_level0, PROV.hadMember):
                self.assertTrue(
                    (output_level1, RDF.type, ALPACA.DataObjectEntity) in prov_graph)
                self.assertTrue(
                    (output_level1, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel1) in prov_graph)
                members = list(prov_graph.objects(output_level1, PROV.hadMember))
                self.assertEqual(len(members), 3)
                for output_level2 in prov_graph.objects(output_level1, PROV.hadMember):
                    self.assertTrue(
                        (output_level2, RDF.type, ALPACA.DataObjectEntity) in prov_graph)
                    self.assertTrue(
                        (output_level2, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel2) in prov_graph)
                    members = list(prov_graph.objects(output_level2, PROV.hadMember))
                    self.assertEqual(len(members), 0)

    def test_provenance_annotation_container_multiple_output_multiple_annotations_range(self):
        activate(clear=True)
        container = process_multiple_container_output_multiple_annotations_range()
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that the annotations exist
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessMultipleContainerOutputMultipleAnnotationsRange)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel2)))
            ), 6)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel1)))
            ), 2)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel0)))
            ), 0)

        # FunctionExecution is ProcessMultipleContainerOutputMultipleAnnotationsRange
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        self.assertTrue((execution_uri, RDF.type,
                         self.ONTOLOGY.ProcessMultipleContainerOutputMultipleAnnotationsRange) in prov_graph)

        # Check returned values
        # First level was skipped, so only classes for Levels 1 and 2 should
        # be present
        output_nodes = prov_graph.subjects(RDF.type,
                                self.ONTOLOGY.ProcessedMultipleContainerOutputLevel1)
        for output_level1 in output_nodes:
            self.assertTrue((output_level1, PROV.wasGeneratedBy, execution_uri) in prov_graph)
            self.assertTrue((output_level1,
                             RDF.type, ALPACA.DataObjectEntity) in prov_graph)
            self.assertTrue((output_level1,
                             RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel1)
                            in prov_graph)
            members = list(prov_graph.objects(output_level1, PROV.hadMember))
            self.assertEqual(len(members), 3)
            for output_level2 in prov_graph.objects(output_level1, PROV.hadMember):
                self.assertTrue(
                    (output_level2, RDF.type, ALPACA.DataObjectEntity) in prov_graph)
                self.assertTrue(
                    (output_level2, RDF.type, self.ONTOLOGY.ProcessedMultipleContainerOutputLevel2) in prov_graph)
                members = list(prov_graph.objects(output_level2, PROV.hadMember))
                self.assertEqual(len(members), 0)

    def test_provenance_annotation_input(self):
        activate(clear=True)
        result = process_input_annotation(5, 6)
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that the annotations exist
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessInputAnnotation)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.Input)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.Param)))
            ), 1)

        # FunctionExecution is ProcessInputAnnotation
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        self.assertTrue((execution_uri, RDF.type,
                         self.ONTOLOGY.ProcessInputAnnotation) in prov_graph)


        # Check input values
        input_nodes = prov_graph.objects(execution_uri, PROV.used)
        for input_node in input_nodes:
            self.assertTrue((execution_uri, PROV.used, input_node) in prov_graph)
            self.assertTrue((input_node,
                             RDF.type, ALPACA.DataObjectEntity) in prov_graph)
            self.assertTrue((input_node,
                             RDF.type, self.ONTOLOGY.Input) in prov_graph)

    def test_provenance_annotation_container_input(self):
        activate(clear=True)
        input_list = [20, 30, 40]
        result = process_container_input_annotation(5, input_list, 6)
        deactivate()

        prov_data = save_provenance()

        # Read PROV information as RDF
        prov_graph = Graph()
        with io.StringIO(prov_data) as data_stream:
            prov_graph.parse(data_stream, format='turtle')

        # Check that the annotations exist
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ProcessContainerInputAnnotation)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.Input)))
            ), 1)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.ContainerElementInput)))
            ), 3)
        self.assertEqual(
            len(list(prov_graph.triples(
                (None, RDF.type, self.ONTOLOGY.Param)))
            ), 1)

        # FunctionExecution is ProcessContainerInputAnnotation
        execution_uri = list(
            prov_graph.subjects(RDF.type, ALPACA.FunctionExecution))[0]
        self.assertTrue((execution_uri, RDF.type,
                         self.ONTOLOGY.ProcessContainerInputAnnotation) in prov_graph)


        # Check input values have the expected number of classes
        input_nodes = prov_graph.objects(execution_uri, PROV.used)
        input_node_count = Counter()
        for input_node in input_nodes:
            self.assertTrue((execution_uri, PROV.used, input_node) in prov_graph)
            for input_node_type in prov_graph.objects(input_node, RDF.type):
                input_node_count[input_node_type] += 1

        self.assertEqual(input_node_count[ALPACA.DataObjectEntity], 4)
        self.assertEqual(input_node_count[self.ONTOLOGY.Input], 1)
        self.assertEqual(input_node_count[self.ONTOLOGY.ContainerElementInput], 3)


"""
This module implements classes to identify Python objects through the analysis
of nodes from an Abstract Syntax Tree, and generates the relationships
between them. The classes support building a hierarchical tree describing
the child/parent relationships between the objects.

Example
-------

For `spiketrains[0]`, the object `spiketrains[0]` is
represented by an `ast.Subscript` AST node. Its information is retrieved by a
`_SubscriptRelationship` object in this module. This will provide the actual
`spiketrains[0]` Python object and an `DataObject` tuple with the index
information (i.e., `0`).

The `spiketrains` container is represented by an `ast.Name` AST node,
whose information is retrieved by the `_NameRelationship` object.
When creating `_NameRelationship`, the `child` parameter will be the
`_SubscriptRelationship` obtained previously for the indexing operation (i.e.,
`spiketrains[0]`).

Therefore, the object tree is created. After the tree is created, the
appropriate `FunctionExecution` named tuples to express those relationships
in the provenance history are created through the `get_function_execution`
method.
"""

import ast
from alpaca.alpaca_types import FunctionExecution, FunctionInfo
import uuid


class _StaticRelationship(object):
    """
    Base class for relationships extracted through static code analysis.

    The information from a single Abstract Syntax Tree node is extracted and
    the reference to the actual Python object that the node represents is
    stored, as well as its information as the `DataObject` named tuple.

    Parameters
    ----------
    node : ast.AST
        Abstract Syntax Tree node that represents the object.
        The subclass varies for each operation.
    time_stamp : datetime
        Time stamp associated with this operation.
    child : _StaticRelationship, optional
        A `_StaticRelationship` object that is owned by the one being created.
        If a node in the Abstract Syntax Tree contains other nodes that
        describe a Python object being tracked, the `_StaticRelationship`
        object obtained from this node is the parent.

        Example: for `spiketrains[0]`, the `_SubscriptRelationship` associated
        with `ast.Subscript` that represents `spiketrains[0]` is the child of
        the `_NameRelationship` associated with the `ast.Name` that represents
        `spiketrains`. In this `ast.Name`, `ast.Subscript` is a value of the
        `
        Default: None.

    Attributes
    ----------
    object_info : alpaca_types.DataObject
        Named tuple describing the Python object associated with this
        `_StaticRelationship` instance.
    parent : _StaticRelationship
        `_StaticRelationship` object that owns this instance.
    value : object
        Reference to the actual Python object associated with this
        `_StaticRelationship` instance.

    Raises
    ------
    TypeError
        If `node` is not of the type describing the operation represented
        by the `_StaticRelationship` object.

    Example
    -------
    For `spiketrains[0]`, a `_SubscriptRelationship` object will have
    :attr:`value` pointing to `spiketrains[0]`, and a :attr:`parent` pointing
    to a `_NameRelationship` object whose :attr:`value` points to
    `spiketrains`. For `_NameRelationship`, :attr:`parent` is None as this is
    the root of the relationships.
    """

    _operation = None
    _node_type = None

    def __init__(self, node, time_stamp, child=None):
        if not isinstance(node, self._node_type):
            raise TypeError("AST node must be of type '"
                            f"{type(self._node_type)}'")
        self.parent = None
        self._node = node
        if child is not None:
            child.set_parent(self)
        self.object_info = None
        self.time_stamp = time_stamp

    def set_parent(self, parent):
        self.parent = parent

    @property
    def value(self):
        """
        Returns the Python object associated with this instance.
        """
        raise NotImplementedError

    def _get_params(self):
        raise NotImplementedError

    def get_function_execution(self):
        """
        Returns a `FunctionExecution` named tuple describing the relationships
        between parent and child nodes. The `params` element will contain the
        relevant information of the relationship (e.g., slice range, index
        number, or attribute name). The function name in the `FunctionInfo`
        named tuple will be the operation name (e.g., 'attribute',
        'subscript', or 'variable'.
        """

        params = self._get_params()
        input_object = self.parent.object_info if self.parent is not None \
            else None
        output_object = self.object_info

        execution_id = str(uuid.uuid4())

        return FunctionExecution(
            function=FunctionInfo(name=self._operation,
                                  module="",
                                  version=""),
            input={0: input_object},
            params=params,
            output={0: output_object},
            arg_map=None,
            kwarg_map=None,
            call_ast=self._node,
            code_statement=None,
            time_stamp_start=self.time_stamp,
            time_stamp_end=self.time_stamp,
            return_targets=[],
            order=None,
            execution_id=execution_id)


class _NameRelationship(_StaticRelationship):
    """
    Static relationship that represents an `ast.Name` Abstract Syntax Tree
    node.

    This is supposed to be the first level of a tree describing the
    dependencies between the objects, and maps to a variable in the script.

    The node must be previously modified to include the reference to the
    Python object associated with the variable, and a `DataObject` named tuple
    with the information from the object.
    """

    _operation = 'variable'
    _node_type = ast.Name

    def __init__(self, node, time_stamp, child=None):
        super(_NameRelationship, self).__init__(node, time_stamp, child)
        self.object_info = node.object_info

    @property
    def value(self):
        return self._node.instance

    def _get_params(self):
        return None


class _SubscriptRelationship(_StaticRelationship):
    """
    Static relationship that represents an `ast.Subscript` Abstract Syntax
    Tree node.

    This represents a subscripting operation in the script.
    """

    _operation = 'subscript'
    _node_type = ast.Subscript

    def __init__(self, node, time_stamp, child):
        super(_SubscriptRelationship, self).__init__(node, time_stamp, child)
        self._slice = self._get_slice(node.slice)

    @staticmethod
    def _get_slice(slice_node):
        # Extracts index or slice information from an `ast.Slice` or
        # `ast.Index` nodes that are the `slice` attribute of `ast.Subscript`.
        # Returns the slice/index value, that will be used to fetch the
        # actual Python object returned by the subscript operation.

        if isinstance(slice_node, ast.Index):

            # Integer or string indexing
            if isinstance(slice_node.value, ast.Num):
                index_value = int(slice_node.value.n)
            elif isinstance(slice_node.value, ast.Str):
                index_value = slice_node.value.s
            elif isinstance(slice_node.value, ast.Name):
                from alpaca.decorator import Provenance
                index_value = Provenance._get_script_variable(slice_node.value.id)
            elif isinstance(slice_node.value, ast.UnaryOp) and \
                isinstance(slice_node.value.op, ast.USub):
                # Negative indexing
                index_value = -int(slice_node.value.operand.n)
            else:
                raise TypeError("Operation not supported")

            return index_value

        # Required for newer Python versions
        if isinstance(slice_node, ast.Constant):
            index_value = slice_node.value
            return index_value

        if isinstance(slice_node, ast.UnaryOp) and \
                isinstance(slice_node.op, ast.USub):
            # Negative indexing
            return -int(slice_node.operand.n)

        if isinstance(slice_node, ast.Name):
            from alpaca.decorator import Provenance
            index_value = Provenance._get_script_variable(slice_node.id)
            return index_value

        if isinstance(slice_node, ast.Slice):
            # Slicing
            stop = getattr(slice_node, 'upper')
            start = getattr(slice_node, 'lower', None)
            step = getattr(slice_node, 'step', None)

            stop = int(stop.n) if stop is not None else None
            start = int(start.n) if start is not None else None
            step = int(step.n) if step is not None else None

            return slice(start, stop, step)

    def _get_params(self):
        params = {}
        if isinstance(self._slice, slice):
            start = self._slice.start
            stop = self._slice.stop
            step = self._slice.step

            params['slice'] = f":{stop}" if stop is not None else ":"
            if start is not None:
                params['slice'] = f"{start}{params['slice']}"
            if step is not None:
                params['slice'] += f":{step}"
        else:
            params['index'] = self._slice

        return params

    @property
    def value(self):
        return self.parent.value[self._slice]


class _AttributeRelationship(_StaticRelationship):
    """
    Static relationship that represents an `ast.Attribute` Abstract Syntax
    Tree node.

    This represents accessing an object attribute using dot '.' operation in
    the script.
    """

    _operation = 'attribute'
    _node_type = ast.Attribute

    def __init__(self, node, time_stamp, child=None):
        super(_AttributeRelationship, self).__init__(node, time_stamp, child)

    def _get_params(self):
        return {'name': self._node.attr}

    @property
    def value(self):
        return getattr(self.parent.value, self._node.attr)



"""
This module contains conversion functions, that will ensure that attributes
and metadata are represented as valid strings according to XSD literals and
other classes of the PROV model (e.g., containers and members).
"""

from alpaca.serialization.neo import _neo_to_prov


__all__ = ['_ensure_type']


def _list_to_prov(value):
    return str(value)


def _quantity_to_prov(value):
    return str(value)


TYPES_MAP = {
    list: _list_to_prov,
}


PACKAGES_MAP = {
    'neo': _neo_to_prov,
    'quantities': _quantity_to_prov,
}


def _ensure_type(value):
    # Function that guarantees that a value of specific types is properly
    # represented as an XSD valid type.
    # Converters can be defined per package (in the `PACKAGES_MAP`
    # dictionary) or type (`TYPES_MAP` dictionary). Base types are not
    # converted, as they are already supported.

    value_type = type(value)
    package = value_type.__module__.split(".")[0]

    if package in PACKAGES_MAP:
        return PACKAGES_MAP[package](value)
    if value_type in TYPES_MAP:
        return TYPES_MAP[value_type](value)
    if isinstance(value, (int, str, bool, float)):
        return value

    # Convert to string by default
    return str(value)
